import socket
import logging as log
from pathlib import Path
from ikpy.chain import Chain
from typing import Optional, Tuple
from .types import DobotError, URDF

class DobotSocketConnection:
    def __init__(self, ip: str, port: int):
        self.socket = socket.socket()
        self.socket.settimeout(10.0)
        self.socket.connect((ip, port))
        log.debug("Connection established")

    " Sends a desired command over the socket connection and returns the potential error and return value in the form of a string "
    " Example return: 0, {}, EnableRobot() -> (None, "") "
    def send_command(self, cmd: str) -> Tuple[Optional[DobotError], str]:
        encoded_cmd = cmd.encode("utf-8")
        self.socket.send(encoded_cmd)
        log.debug(f'The command "{encoded_cmd}" has been sent.')
        return self.__await_reply()

    " This is a quick solution, but may not cover all errors - in which it might panic. "
    def __await_reply(self) -> Tuple[Optional[DobotError], str]:
        data = self.socket.recv(1024)
        response: str = str(data, encoding="utf-8")
        log.debug(f'The return message was "{response}".')
        split_response = response.split(",")
        errorID: int = int(split_response[0].strip())
        return_value: str = split_response[1].strip()
        if errorID == 0:
            return (None, return_value[1:-1])
        else:
            " This will panic if there is no error for the value of errorID, make a pull request. "
            return (DobotError(errorID), return_value[1:-1])
    
    def close(self):
        if self.socket:
            self.socket.close()

    def __del__(self):
        self.close()

class Simulator:
    def __init__(self, fn: URDF) -> None:
        self.chain = Chain.from_urdf_file(fn)

        
def clamp(val: int, local_min: int, local_max: int) -> int:
    log.info(f"{val} was clamped to the range {local_min}, {local_max}")
    return max(local_min, min(val, local_max))

     


import logging as log
from enum import IntEnum
import numpy as np
from dataclasses import dataclass
from strenum import StrEnum


MOVEMENT_PORT = 30003
REALTIME_FEEDBACK_PORT = 30004
DASHBOARD_PORT = 29999


# There could be more, but not documented in:
# https://github.com/Dobot-Arm/TCP-IP-Protocol/blob/master/README-EN.md
class DobotError(IntEnum):
    FAIL_TO_GET = -1
    COMMAND_ERROR = -10000
    PARAMETER_NUM_ERROR = -20000
    WRONG_PARAM_TYPE = -30000
    FIRST_PARAM_INCORRECT = -30001
    SECOND_PARAM_INCORRECT = -30002
    PARAMETER_RANGE_INCORRECT = -40000
    FIRST_PARAM_RANGE = -40001
    SECOND_PARAM_RANGE = -40002

@dataclass
class IOPort:
    mode: int
    distance: int
    index: int
    status: int

    def __post_init__(self, mode: int, distance: int, index: int, status: int):
        self.mode = self.__clamp(mode, 0, 1)
        self.distance = self.__clamp(distance, 0, 100)
        self.index = self.__clamp(index, 1, 24)
        self.status = self.__clamp(status, 0, 1)
    
    def __clamp(self, val: int, local_min: int, local_max: int) -> int:
        log.info(f"{val} was clamped to the range {local_min}, {local_max}")
        return max(local_min, min(val, local_max))

class RobotMode(IntEnum):
    INIT = 1
    BRAKE_OPEN = 2
    RESERVED = 3
    DISABLED = 4
    ENABLE = 5
    BACKDRIVE = 6
    RUNNING = 7
    RECORDING = 8
    ERROR = 9
    PAUSE = 10
    JOG = 11

class JointSelection(StrEnum):
    J1NEG = "j1-"
    J1POS = "j1+"
    J2NEG = "j2-"
    J2POS = "j2+"
    J3NEG = "j3-"
    J3POS = "j3+"
    J4NEG = "j4-"
    J4POS = "j4+"
    J5NEG = "j5-"
    J5POS = "j5+"


class RobotType(IntEnum):
    CR3 = 3
    CR3L = 31
    CR5 = 5
    CR7 = 7
    CR10 = 10
    CR12 = 12
    CR16 = 16
    MG400 = 1
    M1PRO = 2
    NOVA2 = 101
    NOVA5 = 103
    CR3V2 = 113
    CR5V2 = 115
    CR10V2 = 120

class URDF(StrEnum):
    M1PRO = "urdf/m1pro_description.urdf"

FeedbackType = np.dtype([(
    'len',
    np.int64,
), (
    'digital_input_bits',
    np.uint64,
), (
    'digital_output_bits',
    np.uint64,
), (
    'robot_mode',
    np.uint64,
), (
    'time_stamp',
    np.uint64,
), (
    'time_stamp_reserve_bit',
    np.uint64,
), (
    'test_value',
    np.uint64,
), (
    'test_value_keep_bit',
    np.float64,
), (
    'speed_scaling',
    np.float64,
), (
    'linear_momentum_norm',
    np.float64,
), (
    'v_main',
    np.float64,
), (
    'v_robot',
    np.float64,
), (
    'i_robot',
    np.float64,
), (
    'i_robot_keep_bit1',
    np.float64,
), (
    'i_robot_keep_bit2',
    np.float64,
), ('tool_accelerometer_values', np.float64, (3, )),
    ('elbow_position', np.float64, (3, )),
    ('elbow_velocity', np.float64, (3, )),
    ('q_target', np.float64, (6, )),
    ('qd_target', np.float64, (6, )),
    ('qdd_target', np.float64, (6, )),
    ('i_target', np.float64, (6, )),
    ('m_target', np.float64, (6, )),
    ('q_actual', np.float64, (6, )),
    ('qd_actual', np.float64, (6, )),
    ('i_actual', np.float64, (6, )),
    ('actual_TCP_force', np.float64, (6, )),
    ('tool_vector_actual', np.float64, (6, )),
    ('TCP_speed_actual', np.float64, (6, )),
    ('TCP_force', np.float64, (6, )),
    ('Tool_vector_target', np.float64, (6, )),
    ('TCP_speed_target', np.float64, (6, )),
    ('motor_temperatures', np.float64, (6, )),
    ('joint_modes', np.float64, (6, )),
    ('v_actual', np.float64, (6, )),
    # ('dummy', np.float64, (9, 6))])
    ('hand_type', np.byte, (4, )),
    ('user', np.byte,),
    ('tool', np.byte,),
    ('run_queued_cmd', np.byte,),
    ('pause_cmd_flag', np.byte,),
    ('velocity_ratio', np.byte,),
    ('acceleration_ratio', np.byte,),
    ('jerk_ratio', np.byte,),
    ('xyz_velocity_ratio', np.byte,),
    ('r_velocity_ratio', np.byte,),
    ('xyz_acceleration_ratio', np.byte,),
    ('r_acceleration_ratio', np.byte,),
    ('xyz_jerk_ratio', np.byte,),
    ('r_jerk_ratio', np.byte,),
    ('brake_status', np.byte,),
    ('enable_status', np.byte,),
    ('drag_status', np.byte,),
    ('running_status', np.byte,),
    ('error_status', np.byte,),
    ('jog_status', np.byte,),
    ('robot_type', np.byte,),
    ('drag_button_signal', np.byte,),
    ('enable_button_signal', np.byte,),
    ('record_button_signal', np.byte,),
    ('reappear_button_signal', np.byte,),
    ('jaw_button_signal', np.byte,),
    ('six_force_online', np.byte,),
    ('reserve2', np.byte, (82, )),
    ('m_actual', np.float64, (6, )),
    ('load', np.float64,),
    ('center_x', np.float64,),
    ('center_y', np.float64,),
    ('center_z', np.float64,),
    ('user[6]', np.float64, (6, )),
    ('tool[6]', np.float64, (6, )),
    ('trace_index', np.float64,),
    ('six_force_value', np.float64, (6, )),
    ('target_quaternion', np.float64, (4, )),
    ('actual_quaternion', np.float64, (4, )),
    ('reserve3', np.byte, (24, ))])



import logging as log
from typing import Optional
from .util import *
from .types import *


# TODO: The outer dashboard classes will implement all shared functionality, but the CR version will have additional methods.
# TODO: A better naming convention

# We leave it upon the caller of the API to see if they care about the errors.
# There should be no error handling inside of this API.
# This is just a faithful translation and communication layer.


class Dobot:
    def __init__(self, ip: str, urdf_file: Optional[URDF] = None, is_cr: bool = False, logging: bool = False, log_name: str = "output.log", log_level = log.DEBUG):
        if logging:
            log.basicConfig(filename=log_name, level=log_level)
        self.simulator: Optional[Simulator] = Simulator(urdf_file) if urdf_file else None
        self.movement: Movement = Movement(ip)
        self.feedback: Feedback = Feedback(ip)
        self.dashboard: Dashboard = Dashboard(ip)


class Movement(DobotSocketConnection):
    def __init__(self, ip: str):
        super().__init__(ip, MOVEMENT_PORT)

    # MovJ
    def move_joint(
        self, x: float, y: float, z: float, rx: float, ry: float, rz: float
    ) -> Optional[DobotError]:
        opt_error, ret_val = self.send_command(f"MovJ({x}, {y}, {z}, {rx}, {ry}, {rz})")
        return opt_error

    # MovJIO
    def move_joint_io(
        self, x: float, y: float, z: float, rx: float, ry: float, rz: float, io_ports: list[IOPort]
    ) -> Optional[DobotError]:
        command = f"MovJIO({x}, {y}, {z}, {rx}, {ry}, {rz}"
        if io_ports:
            for io_port in io_ports:
                command += f",{{{io_port.mode}, {io_port.distance}, {io_port.index}, {io_port.status}}}" 
        command += ")"
        opt_error, ret_val = self.send_command(command)
        return opt_error

    # JointMovJ
    def joint_to_joint_move(self, joints: list[float]) -> Optional[DobotError]:
        inner = ', '.join(str(joint) for joint in joints)
        command = f"JointMovJ({inner})"
        opt_error, ret_val = self.send_command(command)
        return opt_error

    # MovL
    def move_linear(
        self, x: float, y: float, z: float, rx: float, ry: float, rz: float
    ) -> Optional[DobotError]:
        opt_error, ret_val = self.send_command(f"MovL({x}, {y}, {z}, {rx}, {ry}, {rz})")
        return opt_error

    # MovLIO
    def move_linear_io(
        self, x: float, y: float, z: float, rx: float, ry: float, rz: float, io_ports: list[IOPort]
    ) -> Optional[DobotError]:
        command = f"MovLIO({x}, {y}, {z}, {rx}, {ry}, {rz}"
        if io_ports:
            for io_port in io_ports:
                command += f",{{{io_port.mode}, {io_port.distance}, {io_port.index}, {io_port.status}}}" 
        command += ")"
        opt_error, ret_val = self.send_command(command)
        return opt_error
    
    # Arc
    def move_arc(
        self, x: float, y: float, z: float, rx: float, ry: float, rz:float, x2:float, y2: float, z2: float, rx2: float, ry2: float, rz2:float
    ) -> Optional[DobotError]:
        opt_error, ret_val = self.send_command(f"Arc({x}, {y}, {z}, {rx}, {ry}, {rz}, {x2}, {y2}, {z2}, {rx2}, {ry2}, {rz2})")
        return opt_error
    
    # Sync
    def sync(self) -> Optional[DobotError]:
        opt_error, ret_val = self.send_command("Sync()")
        return opt_error
    
    # RelMovJUser
    def relative_move_joint(self, offx: float, offy: float, offz: float, offrx: float, offry: float, offrz: float, user_index: int) -> Optional[DobotError]:
        user_index = clamp(user_index, 0, 9)
        opt_error, ret_val = self.send_command(f"RelMovJUser({offx}, {offy}, {offz}, {offrx}, {offry}, {offrz}, {user_index})")
        return opt_error

    # RelMovLUser
    def relative_linear_joint(self, offx: float, offy: float, offz: float, offrx: float, offry: float, offrz: float, user_index: int) -> Optional[DobotError]:
        user_index = clamp(user_index, 0, 9)
        opt_error, ret_val = self.send_command(f"RelMovLUser({offx}, {offy}, {offz}, {offrx}, {offry}, {offrz}, {user_index})")
        return opt_error

    # RelJointMovJ
    def relative_joint_motion(self, off1: float, off2: float, off3: float, off4: float, off5: float, off6: float) -> Optional[DobotError]:
        opt_error, ret_val = self.send_command(f"RelJointMovJ({off1}, {off2}, {off3}, {off4}, {off5}, {off6})")
        return opt_error

    # MoveJog
    def move_jog(self, joint: JointSelection) -> Optional[DobotError]:
        opt_error, ret_val = self.send_command(f"MovJog({joint})")
        return opt_error


class Dashboard(DobotSocketConnection):
    def __init__(self, ip: str):
        super().__init__(ip, DASHBOARD_PORT)

    def turn_on(self) -> Optional[DobotError]:
        opt_error, ret_val = self.send_command("PowerOn()")
        return opt_error

    def enable(self) -> Optional[DobotError]:
        opt_error, ret_val = self.send_command("EnableRobot()")
        return opt_error

    def disable(self) -> Optional[DobotError]:
        opt_error, ret_val = self.send_command("DisableRobot()")
        return opt_error

    def reset(self) -> Optional[DobotError]:
        opt_error, ret_val = self.send_command("ResetRobot()")
        return opt_error

    def clear_errors(self) -> Optional[DobotError]:
        opt_error, ret_val = self.send_command("ClearError()")
        return opt_error

    def emergency_stop(self) -> Optional[DobotError]:
        opt_error, ret_val = self.send_command("EmergencyStop()")
        return opt_error

    # TODO: Cover all possible Error IDs
    def get_error_id(self):
        opt_error, ret_val = self.send_command("GetErrorID()")
        log.info(f"Error IDs : {ret_val}")
        return opt_error

    def get_digital_input(self, index: int) -> DobotError | int:
        index = clamp(index, 1, 32)
        opt_error, ret_val = self.send_command(f"DI({index})")
        try:
            return int(ret_val)
        except:
            return opt_error

    def set_digital_output(self, index: int, val: int) -> Optional[DobotError]:
        index = clamp(index, 1, 16)
        val = clamp(val, 0, 1)
        opt_error, ret_val = self.send_command(f"DO({index}, {val})")
        return opt_error
    
    def robot_mode(self) -> DobotError | RobotMode:
        opt_error, ret_val = self.send_command("RobotMode()")
        if opt_error:
            return opt_error
        else:
            # NOTE: If this fails, then we have an uncovered error.
            return RobotMode(int(ret_val))
    
    def set_linear_accel(self, rate: int) -> Optional[DobotError]:
        rate = clamp(rate, 1, 100)
        opt_error, ret_val = self.send_command(f"AccL({rate})")
        return opt_error

    def set_joint_accel(self, rate: int) -> Optional[DobotError]:
        rate = clamp(rate, 1, 100)
        opt_error, ret_val = self.send_command(f"AccJ({rate})")
        return opt_error
    
    def set_linear_velocity(self, rate: int) -> Optional[DobotError]:
        rate = clamp(rate, 1, 100)
        opt_error, ret_val = self.send_command(f"SpeedL({rate})")
        return opt_error

    def set_joint_velocity(self, rate: int) -> Optional[DobotError]:
        rate = clamp(rate, 1, 100)
        opt_error, ret_val = self.send_command(f"SpeedJ({rate})")
        return opt_error

    def set_speedfactor(self, ratio: int) -> Optional[DobotError]:
        ratio = clamp(ratio, 1, 100)
        opt_error, ret_val = self.send_command(f"SpeedFactor({ratio})")
        return opt_error
    
    def set_arc_params(self, index: int) -> Optional[DobotError]:
        index = clamp(index, 0, 9)
        opt_error, ret_val = self.send_command(f"Arch({index})")
        return opt_error

    def set_continuous_path(self, ratio: int) -> Optional[DobotError]:
        ratio = clamp(ratio, 1, 100)
        opt_error, ret_val = self.send_command(f"CP({ratio})")
        return opt_error

    def set_user(self, index: int) -> Optional[DobotError]:
        index = clamp(index, 0, 9)
        opt_error, ret_val = self.send_command(f"User({index})")
        return opt_error

    def set_tool(self, index: int) -> Optional[DobotError]:
        index = clamp(index, 0, 9)
        opt_error, ret_val = self.send_command(f"Tool({index})")
        return opt_error
    
    def set_payload(self, weight: float, inertia: float) -> Optional[DobotError]:
        # TODO: Figure out acceptable ranges of weights and inertias for each model.
        opt_error, ret_val = self.send_command(f"PayLoad({weight, inertia})")
        return opt_error
    
    def set_digital_output(self, index: int, status: int) -> Optional[DobotError]:
        status = clamp(status, 0, 1)
        opt_error, ret_val = self.send_command(f"DO({index}, {status})")
        return opt_error
    
    def run_script(self, name: str) -> Optional[DobotError]:
        opt_error, ret_val = self.send_command(f"RunScript({name})")
        return opt_error
    
    def stop_script(self) -> Optional[DobotError]:
        opt_error, ret_val = self.send_command("StopScript()")
        return opt_error

    def pause_script(self) -> Optional[DobotError]:
        opt_error, ret_val = self.send_command("PauseScript()")
        return opt_error

    def continue_script(self) -> Optional[DobotError]:
        opt_error, ret_val = self.send_command("ContinueScript()")
        return opt_error


# TODO: Create a numpy type which houses all needed values
# NOTE: Different for CR version
   
class Feedback(DobotSocketConnection):
    def __init__(self, ip: str):
        super().__init__(ip, REALTIME_FEEDBACK_PORT)
    
